import sys
import biom
import pandas as pd
from qiime2 import Artifact


def per_sample_filter(feature_table, filtering_integer):
    '''
    Filters reads out of a QIIME2 feature table according to a minimum read count requirement *per sample*

    feature_table    : path to QIIME2 feature table

    filtering_integer: Either an integer for even filtering across samples, or path to the
                       Expected_False_Reads_Per_Index.csv output by index_jump.py

    Return: Frequency filtered QIIME2 feature table of type FeatureTable[Frequency]

    Output: 'freq_filt_table.qza' QIIME2 artifact of type FeatureTable[Frequency]
    '''

    # Importing from a biom table
    try:
        working_table = biom.load_table(feature_table)
    except:
        print('Could not import from biom table, trying from qza...')
        try:
            # Importing from imported qza
            qza_table = Artifact.load(feature_table)
            working_table = qza_table.view(biom.Table)
        except:
            print('Could not import from qiime artifact, check to ensure input file is appropriate format')
            exit()

    # Convert table to df
    biom_df = working_table.to_dataframe()

    # Filtering if integer is specified
    if type(filtering_integer) == int:

        print('File loaded! Filtering at '+str(filtering_integer)+' occurrences per feature')

        # Filter out any feature from a sample if it has less than X reads,
        biom_df = biom_df._get_numeric_data()
        biom_dense = biom_df.sparse.to_dense()

        biom_dense[biom_dense < filtering_integer] = 0

        # Translate the filtered table back into a qza
        biom_dense = biom_dense.transpose()
        freq_filt_table = Artifact.import_data("FeatureTable[Frequency]", biom_dense)

        # Export the new table as a qza for further use in Qiime2
        freq_filt_table.save('freq_filt_table.qza')

        return freq_filt_table

    # Filtering based on per-sample CSV
    else:
        print(f'File loaded! Filtering according to per-sample values specified in {filtering_integer}')
        per_sample_table = pd.read_csv(filtering_integer, index_col='SampleIndex')
        per_sample_table_trans = per_sample_table.transpose()

        biom_df = biom_df._get_numeric_data()
        biom_dense = biom_df.sparse.to_dense()

        for k, v in per_sample_table_trans.iteritems():
            for column, values in biom_dense.iteritems():
                if k == column:
                    working_values = values.tolist()
                    edited_values = [0 if x <= int(v) else x for x in working_values]
                    biom_dense[column] = edited_values

        # Translate the filtered table back into a qza
        biom_dense = biom_dense.transpose()
        freq_filt_table = Artifact.import_data("FeatureTable[Frequency]", biom_dense)

        # Export the new table as a qza for further use in Qiime2
        freq_filt_table.save('freq_filt_table.qza')

        return freq_filt_table


def main():
    if sys.argv[2].endswith('.csv'):
        per_sample_filter(sys.argv[1], sys.argv[2])
    else:
        per_sample_filter(sys.argv[1], int(sys.argv[2]))


if __name__ == '__main__':
    main()



import os
import subprocess
import pandas as pd
import sys


def remultiplex(sequenceFile, indexFile):
    '''
    Takes dual-indexed reads, trims the 5' and 3' ends of the reads past the indexes, and moves the 3' index to
    immediately follow the 5' index (i.e. ['MultiplexedSingleEndBarcodeInSequence'] format)

    sequenceFile : path to raw sequence file of type .fastq, .fastq.gz, or .bam

    indexFile    : path to .csv containing all the index tag sequences that are present in the sequencing pool. This .csv
                   should be formatted as specified below (See indexes.csv for reference)

    Return : None

    Output : remultiplexed_seqs.fastq.gz
    '''
    if sequenceFile.endswith('.bam'):
        convert = str('samtools fastq '+sequenceFile+' > raw_seqs.fastq')
        subprocess.run(convert, shell=True)
        sequenceFile = './raw_seqs.fastq'

    else:
        assert sequenceFile.endswith('.fastq') or sequenceFile.endswith('.gz'), 'Sequence file not of proper format. Should be .bam or .fastq'

    # Read in indexes csv, and convert Forward indexes to dictionary
    df = pd.read_csv(indexFile, dtype=str)
    df['ID_ori'] = (df['orientation'] + df['ID'])
    f_df = df[df['orientation'] == 'F']
    fwd_indexes = dict(zip(f_df.ID_ori, f_df.seq))

    # Organizing by Forward Index
    f = open('fwd_sort.txt', 'w')
    for k in fwd_indexes:
        fwd_sort = str('cutadapt -g '+fwd_indexes[k]+' --discard-untrimmed -e 0 -j 0 --overlap 10 -o '
                   +k+'.fastq '+sequenceFile)
        subprocess.run(fwd_sort, shell=True, stdout=f)

    os.system('find . -type f -size 0 -delete')

    # Convert Reverse barcodes to dictionary
    r_df = df[df['orientation'] == 'R']
    rev_indexes = dict(zip(r_df.ID_ori, r_df.seq))

    fastqs = []
    for file in os.listdir('.'):
        if len(file) == 9 and file.endswith('fastq'):
            fastqs.append(file)

    # Iterating over each Forward Index to then organize by Reverse index
    f = open('rev_sort.txt', 'w')
    for fastq in fastqs:
        for k in rev_indexes:
            rev_sort = str('cutadapt -a '+rev_indexes[k]+'$ --discard-untrimmed -e 0 -j 0 --overlap 10 -o '
                           +fastq[0:3]+k+'.fastq '+fastq)
            subprocess.run(rev_sort, shell=True, stdout=f)

    os.system('find . -type f -size 0 -delete')

    # Create dictionary for merged Fwd and Rev idnexes
    merged_indexes = {}
    for k_a, v_a in fwd_indexes.items():
        for k_b, v_b in rev_indexes.items():
            merged_indexes[k_a + k_b] = v_a + v_b

    fastqs = []
    for fastq in os.listdir('.'):
        if len(fastq) == 12 and fastq.endswith('fastq'):
            fastqs.append(fastq)

    # Prepending each read with the appropriate merged index as indicated by file name
    fastqs.sort()
    for fastq in fastqs:
        for k, v in merged_indexes.items():
            if fastq[0:6] == k:
                with open(fastq, 'r') as f, open('remultiplexed_seqs.fastq', 'a+') as out:
                    count = 0
                    for idx, line in enumerate(f.read().splitlines()):
                        count += 1
                        if count == 2:
                            print(v+line, file=out)
                        elif count == 4:
                            print(''.join(['I']*len(v))+line, file=out)
                        else:
                            print(line, file=out)
                        if count == 4:
                            count = 0
            else:
                continue

    os.system('gzip remultiplexed_seqs.fastq')

    os.system('mkdir fastqs')
    os.system('mv *.fastq fastqs| mv fwd_sort.txt fastqs| mv rev_sort.txt fastqs')


def main():
    remultiplex(sys.argv[1], sys.argv[2])


if __name__ == '__main__':
    main()



import os
import sys
import pathlib
import shutil
import tempfile

import pandas as pd
from Bio import SeqIO
from qiime2 import Artifact, Metadata
from qiime2.plugins.feature_table.methods import filter_features, filter_seqs


# Length filter of rep_seqs
def length_filter(feature_table, representative_sequences, length_to_filter):
    '''
    Length filter of QIIME2 feature table and representative sequences

    feature_table: path to QIIME2 qza file of data type FeatureTable[Frequency]

    representative_sequences: path to QIIME2 qza file of data type FeatureData[Sequence]

    length_to_filter: An integer threshold for sequence length. All sequences shorter than specified
                      length will be removed.

    Return: Frequency filtered QIIME2 feature table of type FeatureTable[Frequency]
            Frequency filtered QIIME2 feature table of type FeatureData[Sequence]

    Output: 'length_filt_table.qza' QIIME2 artifact of type FeatureTable[Frequency]
            'length_filt_seqs.qza' QIIME2 artifact of type FeatureData[Sequence]
    '''

    # Pull fasta from rep_seqs to use as metadata
    def extract_fasta(file, dest):
        with tempfile.TemporaryDirectory() as temp:
            file.export_data(temp)
            temp_pathlib = pathlib.Path(temp)
            for item in temp_pathlib.iterdir():
                if item.suffix == '.fasta':
                    shutil.copy(item, dest)

    # QIIME Artifact Loading and error checking
    try:
        qza_table = Artifact.load(feature_table)
    except:
        print('Could not import feature table, check to ensure input file is appropriate format')
        exit()
    try:
        rep_seqs = Artifact.load(representative_sequences)
    except:
        print('Could not import sequences, check to ensure input file is appropriate format')
        exit()

    print(f'QIIME Artifacts loaded! Filtering reads at a length of {length_to_filter}bp')

    os.makedirs('fastas', exist_ok=True)
    extract_fasta(rep_seqs, 'fastas')

    # Parse fasta to get IDs and lengths, then send to df and filter IDs by length
    with open('fastas/dna-sequences.fasta') as fasta_file:
        featureids = []
        lengths = []
        for seq_record in SeqIO.parse(fasta_file, 'fasta'):
            featureids.append(seq_record.id)
            lengths.append(len(seq_record.seq))

    rep_seqs_meta = pd.DataFrame({"FeatureID": featureids, "Length": lengths})

    features_to_exclude = rep_seqs_meta[rep_seqs_meta['Length'] < length_to_filter]

    features_to_exclude['FeatureID'].to_csv('Features-to-exclude.csv', index=False)

    exclude = Metadata.load("Features-to-exclude.csv")

    shutil.rmtree('fastas')

    # Filter rep-seqs based on seqs_to_exclude
    rep_seqs_filt = filter_seqs(rep_seqs,
                               metadata = exclude,
                               exclude_ids = True)

    # Filter table based on seqs_to_exclude
    table_filt = filter_features(qza_table,
                                metadata = exclude,
                                exclude_ids = True)

    rep_seqs_filt.filtered_data.save(f'length_filt_seqs_{length_to_filter}.qza')
    table_filt.filtered_table.save(f'length_filt_table_{length_to_filter}.qza')

    return table_filt, rep_seqs_filt


def main():
    length_filter(sys.argv[1], sys.argv[2], int(sys.argv[3]))


if __name__ == '__main__':
    main()




import pathlib
import math
import sys
import shutil
import tempfile
import os
import pandas as pd
from qiime2 import Artifact
from qiime2.plugins.demux.visualizers import summarize


def calculate(demultiplexed_seqs, sample_map, calibrator_tag_pairs):
    '''
    Calculate Index Jump Rate based off calibrator Tags

    demultiplexed_seqs: path to demultiplexed QIIME2 qza file of data type SampleData[SequencesWithQuality]

    sample_map: path to tab delimited QIIME2 sample map file

    calibrator_tag_pairs: pairs of calibrator tags, input as either a tuple, or list of tuples, with
                          each index being a 2 digit zero padded string.
                          ex: [('01', '11')] or [('01', '11'), ('02','12')]

    Return: Integer value of maximum number of index jumps (false reads) expected in a single sample
    Output: Expected_False_Reads_Per_Index.csv file containing number of false reads expected in EACH sample
            log.txt file containing summary statistics
    '''

    def extract_tsvs(viz, dest):
        '''
        Extracts useful TSV from QIIME artifact files
        '''
        with tempfile.TemporaryDirectory() as temp:
            viz.export_data(temp)
            temp_pathlib = pathlib.Path(temp)
            for file in temp_pathlib.iterdir():
                if file.suffix == '.tsv':
                    shutil.copy(file, dest)

    # Load in sample_map as pandas df
    meta_df = pd.read_csv(sample_map, sep='\t')

    # Load in demultiplexed sequences from path and Summarize
    demux = Artifact.load(demultiplexed_seqs)
    print('Demultiplexed sequences loaded... Extracting per sample summaries to tsvs directory \n . \n . \n .')
    demux_summary = summarize(demux)
    demux_viz = demux_summary.visualization

    # Make directory, extract tsvs, read in per-sample tsv, set as df
    os.makedirs('tsvs', exist_ok=True)
    extract_tsvs(demux_viz, 'tsvs')
    per_sample_df = pd.read_csv("./tsvs/per-sample-fastq-counts.tsv", sep="\t")
    df = per_sample_df

    # Establishing variables and altering dataframes
    # Create FWD index column
    df['Fwd']     = df['sample ID'].str[1:3]

    # Create REV index column
    df['Rev']     = df['sample ID'].str[4:6]
    df            = df.sort_values(['sample ID'])
    df.reset_index(inplace=True)

    # Add in True_False coding
    df['True_False'] = meta_df['True_False']

    if calibrator_tag_pairs is None:
        print('No calibrator tags given. Calculating index jump rate based on 0s in Sample Map.\n '
              '*NOTE* This calculation method is less accurate than when using calibrator tags!\n'
              'Jump rate estimates will be lower than the true rate and may result in retaining\n'
              'false reads within the sample pool.')

        all_false_reads = df[df["True_False"] == 0]
        summed_false = all_false_reads['forward sequence count'].sum()
        jump_rate = (summed_false/(df['forward sequence count'].sum()))*2

    else:
        # Error flagging for Calibrator Tag input: Scripting / Jupyter-Notebook
        print('Validating Calibrator Tag input... \n . \n . \n .')
        if not all(isinstance(item, tuple) for item in calibrator_tag_pairs):
            raise TypeError(
                'Calibrator Tag Index identifiers are not input in proper format. Should be List of Tuples.')

        # Error flagging for Calibrator Tag input: Command Line input
        for i, j in calibrator_tag_pairs:
            if type(i) != str or type(j) != str:
                raise TypeError('Calibrator Tag Index identifiers are not of proper type. Should be type String.')

            if len(i) != 2 or len(j) != 2:
                raise ValueError('Calibrator Tag Index identifiers are not of proper length. Should be 2 digit inputs.')

        # Calculating the jump_rate of the sequencing run based off of indicated calibrator tags
        individual_jump_rates = []
        for fwd, rev in calibrator_tag_pairs:

            fwd_df          = df[df['Fwd'] == fwd]
            rev_df          = df[df['Rev'] == rev]

            # Validity check for calibrator tags
            fwd_True       = fwd_df[fwd_df['True_False'] == 1]
            assert fwd_True['True_False'].sum() == 1, \
                                                      f'Specified tag pair {fwd, rev} does not meet requirements for ' \
                                                      f'usage as calibrator tags. Should only have one true sequence ' \
                                                      f'specified in sample map.'
            rev_True       = rev_df[rev_df['True_False'] == 1]
            assert rev_True['True_False'].sum() == 1, \
                                                      f'Specified tag pair {fwd, rev} does not meet requirements for ' \
                                                      f'usage as calibrator tags. Should only have one true sequence ' \
                                                      f'specified in sample map.'

            print(f'All checks passed! Calculating Index Jump rate based off calibrator tags {fwd, rev}')
            # Calculations
            fwd_False       = fwd_df[fwd_df['True_False'] == 0]
            fwd_jump_rate   = (fwd_False['forward sequence count'].sum())/(fwd_df['forward sequence count'].sum())
            rev_False       = rev_df[rev_df['True_False'] == 0]
            rev_jump_rate   = (rev_False['forward sequence count'].sum())/(rev_df['forward sequence count'].sum())

            individual_jump_rates.append(fwd_jump_rate)
            individual_jump_rates.append(rev_jump_rate)

        # Calculate Average Jump rate for ALL calibrator tags
        jump_series = pd.Series(individual_jump_rates)
        jump_rate = jump_series.sum()/len(jump_series)

    # Preparing lists for calculations
    Indexes       = df['sample ID'].tolist()
    Seq_counts    = df['forward sequence count'].tolist()

    Fwds          = df['Fwd'].tolist()
    Fwds          = list(set(Fwds))
    Fwds.sort()

    Revs          = df['Rev'].tolist()
    Revs          = list(set(Revs))
    Revs.sort()

    # Calculating each Indexes expected # of False reads
    indexes_exp_false = []
    total_read_count = df['forward sequence count'].sum()

    count = 0
    for i in Fwds:
        temp_df = df[df['Fwd'] == i]

        # No. of reads with i FWD index
        i_seqs = temp_df['forward sequence count'].sum()
        i_seqs_no_j = i_seqs - Seq_counts[count]

        for j in Revs:
            temp_df2 = df[df['Rev'] == j]
            # No. of reads with j REV index
            j_seqs = temp_df2['forward sequence count'].sum()
            j_seqs_no_i = j_seqs-Seq_counts[count]

            # No. of j reads with an expected false i index
            i_jumps = i_seqs * jump_rate
            i_to_j = i_jumps * (j_seqs_no_i / total_read_count)

            # No. of i reads with an expected false j index
            j_jumps = j_seqs * jump_rate
            j_to_i = j_jumps * (i_seqs_no_j / total_read_count)

            # Sum for total i-j pair false read count
            indiv_tag_jump_count = math.ceil((i_to_j + j_to_i))

            # Append to list for data table output
            to_add = (Indexes[count], indiv_tag_jump_count)
            indexes_exp_false.append(to_add)
            count += 1

    # Export csv with expected false reads per sample
    out_df = pd.DataFrame(indexes_exp_false, columns=['SampleIndex', 'Expected False Reads'])
    print('Exporting recommended per-sample filtering levels to Expected_False_Reads_Per_Index.csv')
    out_df.to_csv('Expected_False_Reads_Per_Index.csv', index=False)

    # Text output to log.txt of some summary statistics
    max_IJR = out_df['Expected False Reads'].max()
    false_read_count = out_df['Expected False Reads'].sum()

    with open('log.txt', 'w') as out:
        print(f'Calculated Average jump rate: {jump_rate}', file=out)
        print(f'Total number of false reads: {false_read_count} / {total_read_count}', file=out)
        print(f'Total percent of false reads: {(false_read_count/total_read_count)*100:.3f}%', file=out)
        print(f'Maximum number of false reads expected in a single sample: {max_IJR}', file=out)

    return max_IJR


def main():
    if len(sys.argv) < 4:
        calculate(sys.argv[1], sys.argv[2], calibrator_tag_pairs=None)

    else:
        c_tags_list = []
        for i in range(3, len(sys.argv)):
            c_tag_pair = (tuple(sys.argv[i].split(',')))
            c_tags_list.append(c_tag_pair)
            calculate(sys.argv[1], sys.argv[2], c_tags_list)


if __name__ == '__main__':
    main()



from setuptools import setup

setup(
    name='Metaplex',
    version='1.1.0',
    description='Read Processing and Quality Control Toolkit for Dual-Indexed Metabarcoding',
    url='https://github.com/NGabry/MetaPlex',
    author='Nick Gabry',
    author_email='n.t.gabry@gmail.com',
    license='BSD 3-clause',
    packages=['metaplex'],
    entry_points={
        'console_scripts': [
            'Metaplex-remultiplex=metaplex.remultiplexing:main',
            'Metaplex-calculate-IJR=metaplex.index_jump:main',
            'Metaplex-per-sample-filter=metaplex.per_sample_filtering:main',
            'Metaplex-length-filter=metaplex.length_filtering:main',
        ],
    },
    install_requires=['numpy',
                      'pandas',
                      'cutadapt',
                      'biom-format',
                      'biopython',
                      ],

    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Scientific/Engineering :: Bio-Informatics',

    ],
)
